@@ -6,8 +6,6 @@
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 #
@@ -33,8 +31,6 @@ equivalent code::
 __docformat__ = 'restructuredtext'
 
 from dbus.mainloop.glib import DBusGMainLoop, threads_init
-import _dbus_glib_bindings
-import _dbus_bindings
 
 init_threads = threads_init
 
