@@ -7,8 +7,6 @@
 #
 # This program is free software; you can redistribute it and/or modify
 # it under the terms of the GNU General Public License as published by
-# This program is free software; you can redistribute it and/or modify
-# it under the terms of the GNU General Public License as published by
 # the Free Software Foundation; either version 2 of the License, or
 # (at your option) any later version.
 #
@@ -28,13 +26,20 @@ import sys
 import logging
 import operator
 import traceback
+try:
+    import thread
+except ImportError:
+    import dummy_thread as thread
 
 import _dbus_bindings
-import dbus._dbus as _dbus
-from dbus.exceptions import NameExistsException
-from dbus.exceptions import UnknownMethodException
-from dbus.decorators import method
-from dbus.decorators import signal
+from dbus import SessionBus, Signature, Struct, validate_bus_name, \
+                 validate_object_path, INTROSPECTABLE_IFACE, ObjectPath
+from dbus.decorators import method, signal
+from dbus.exceptions import DBusException, \
+                            NameExistsException, \
+                            UnknownMethodException
+from dbus.lowlevel import ErrorMessage, MethodReturnMessage
+from dbus.proxies import LOCAL_PATH
 
 
 _logger = logging.getLogger('dbus.service')
@@ -78,8 +83,13 @@ class BusName(object):
             `name` : str
                 The well-known name to be advertised
             `bus` : dbus.Bus
-                A Bus on which this service will be advertised; if None
-                (default) the session bus will be used
+                A Bus on which this service will be advertised.
+
+                Omitting this parameter or setting it to None has been
+                deprecated since version 0.82.1. For backwards compatibility,
+                if this is done, the global shared connection to the session
+                bus will be used.
+
             `allow_replacement` : bool
                 If True, other processes trying to claim the same well-known
                 name will take precedence over this one.
@@ -91,9 +101,15 @@ class BusName(object):
                 services waiting for the requested name if another service
                 already holds it.
         """
-        # get default bus
-        if bus == None:
-            bus = _dbus.Bus()
+        validate_bus_name(name, allow_well_known=True, allow_unique=False)
+
+        # if necessary, get default bus (deprecated)
+        if bus is None:
+            import warnings
+            warnings.warn('Omitting the "bus" parameter to '
+                          'dbus.service.BusName.__init__ is deprecated',
+                          DeprecationWarning, stacklevel=2)
+            bus = SessionBus()
 
         # see if this name is already defined, return it if so
         # FIXME: accessing internals of Bus
@@ -101,7 +117,7 @@ class BusName(object):
             return bus._bus_names[name]
 
         # otherwise register the name
-	name_flags = (
+        name_flags = (
             (allow_replacement and _dbus_bindings.NAME_FLAG_ALLOW_REPLACEMENT or 0) |
             (replace_existing and _dbus_bindings.NAME_FLAG_REPLACE_EXISTING or 0) |
             (do_not_queue and _dbus_bindings.NAME_FLAG_DO_NOT_QUEUE or 0))
@@ -226,13 +242,12 @@ def _method_lookup(self, method_name, dbus_interface):
             raise UnknownMethodException('%s is not a valid method' % method_name)
 
 
-# FIXME: if signature is '', we used to accept anything, but now insist on
-# zero args. Which was desired? -smcv
 def _method_reply_return(connection, message, method_name, signature, *retval):
-    reply = _dbus_bindings.MethodReturnMessage(message)
+    reply = MethodReturnMessage(message)
     try:
         reply.append(signature=signature, *retval)
     except Exception, e:
+        logging.basicConfig()
         if signature is None:
             try:
                 signature = reply.guess_signature(retval) + ' (guessed)'
@@ -248,15 +263,17 @@ def _method_reply_return(connection, message, method_name, signature, *retval):
 
 
 def _method_reply_error(connection, message, exception):
-    if hasattr(exception, '_dbus_error_name'):
-        name = exception._dbus_error_name
+    name = getattr(exception, '_dbus_error_name', None)
+
+    if name is not None:
+        pass
     elif getattr(exception, '__module__', '') in ('', '__main__'):
         name = 'org.freedesktop.DBus.Python.%s' % exception.__class__.__name__
     else:
         name = 'org.freedesktop.DBus.Python.%s.%s' % (exception.__module__, exception.__class__.__name__)
 
     contents = traceback.format_exc()
-    reply = _dbus_bindings.ErrorMessage(message, name, contents)
+    reply = ErrorMessage(message, name, contents)
 
     connection.send_message(reply)
 
@@ -295,15 +312,15 @@ class InterfaceType(type):
             # convert signature into a tuple so length refers to number of
             # types, not number of characters. the length is checked by
             # the decorator to make sure it matches the length of args.
-            in_sig = tuple(_dbus_bindings.Signature(func._dbus_in_signature))
+            in_sig = tuple(Signature(func._dbus_in_signature))
         else:
             # magic iterator which returns as many v's as we need
             in_sig = _VariantSignature()
 
         if func._dbus_out_signature:
-            out_sig = _dbus_bindings.Signature(func._dbus_out_signature)
+            out_sig = Signature(func._dbus_out_signature)
         else:
-            # its tempting to default to _dbus_bindings.Signature('v'), but
+            # its tempting to default to Signature('v'), but
             # for methods that return nothing, providing incorrect
             # introspection data is worse than providing none at all
             out_sig = []
@@ -323,7 +340,7 @@ class InterfaceType(type):
         if func._dbus_signature:
             # convert signature into a tuple so length refers to number of
             # types, not number of characters
-            sig = tuple(_dbus_bindings.Signature(func._dbus_signature))
+            sig = tuple(Signature(func._dbus_signature))
         else:
             # magic iterator which returns as many v's as we need
             sig = _VariantSignature()
@@ -338,6 +355,10 @@ class InterfaceType(type):
 class Interface(object):
     __metaclass__ = InterfaceType
 
+#: A unique object used as the value of Object._object_path and
+#: Object._connection if it's actually in more than one place
+_MANY = object()
+
 class Object(Interface):
     r"""A base class for exporting your own Objects across the Bus.
 
@@ -370,58 +391,240 @@ class Object(Interface):
                 return self._last_input
     """
 
-    # the signature of __init__ is a bit mad, for backwards compatibility
+    #: If True, this object can be made available at more than one object path.
+    #: If True but `SUPPORTS_MULTIPLE_CONNECTIONS` is False, the object may
+    #: handle more than one object path, but they must all be on the same
+    #: connection.
+    SUPPORTS_MULTIPLE_OBJECT_PATHS = False
+
+    #: If True, this object can be made available on more than one connection.
+    #: If True but `SUPPORTS_MULTIPLE_OBJECT_PATHS` is False, the object must
+    #: have the same object path on all its connections.
+    SUPPORTS_MULTIPLE_CONNECTIONS = False
+
     def __init__(self, conn=None, object_path=None, bus_name=None):
         """Constructor. Either conn or bus_name is required; object_path
         is also required.
 
         :Parameters:
-            `conn` : dbus.Connection
+            `conn` : dbus.connection.Connection or None
                 The connection on which to export this object.
 
-                If None, use the Bus associated with the given ``bus_name``,
-                or raise TypeError if there is no ``bus_name`` either.
+                If None, use the Bus associated with the given ``bus_name``.
+                If there is no ``bus_name`` either, the object is not
+                initially available on any Connection.
 
                 For backwards compatibility, if an instance of
                 dbus.service.BusName is passed as the first parameter,
                 this is equivalent to passing its associated Bus as
                 ``conn``, and passing the BusName itself as ``bus_name``.
 
-            `object_path` : str
-                The D-Bus object path at which to export this Object.
+            `object_path` : str or None
+                A D-Bus object path at which to make this Object available
+                immediately. If this is not None, a `conn` or `bus_name` must
+                also be provided.
 
-            `bus_name` : dbus.service.BusName
+            `bus_name` : dbus.service.BusName or None
                 Represents a well-known name claimed by this process. A
                 reference to the BusName object will be held by this
                 Object, preventing the name from being released during this
                 Object's lifetime (unless it's released manually).
         """
-        if object_path is None:
-            raise TypeError('The object_path argument is required')
+        if object_path is not None:
+            validate_object_path(object_path)
 
         if isinstance(conn, BusName):
             # someone's using the old API; don't gratuitously break them
             bus_name = conn
             conn = bus_name.get_bus()
         elif conn is None:
-            # someone's using the old API but naming arguments, probably
-            if bus_name is None:
-                raise TypeError('Either conn or bus_name is required')
-            conn = bus_name.get_bus()
+            if bus_name is not None:
+                # someone's using the old API but naming arguments, probably
+                conn = bus_name.get_bus()
+
+        #: Either an object path, None or _MANY
+        self._object_path = None
+        #: Either a dbus.connection.Connection, None or _MANY
+        self._connection = None
+        #: A list of tuples (Connection, object path, False) where the False
+        #: is for future expansion (to support fallback paths)
+        self._locations = []
+        #: Lock protecting `_locations`, `_connection` and `_object_path`
+        self._locations_lock = thread.allocate_lock()
+
+        #: True if this is a fallback object handling a whole subtree.
+        self._fallback = False
 
-        self._object_path = object_path
         self._name = bus_name
-        self._bus = conn
-            
-        self._connection = self._bus.get_connection()
 
-        self._connection._register_object_path(object_path, self._message_cb, self._unregister_cb)
+        if conn is None and object_path is not None:
+            raise TypeError('If object_path is given, either conn or bus_name '
+                            'is required')
+        if conn is not None and object_path is not None:
+            self.add_to_connection(conn, object_path)
 
-    __dbus_object_path__ = property(lambda self: self._object_path, None, None,
-                                    "The D-Bus object path of this object")
+    @property
+    def __dbus_object_path__(self):
+        """The object-path at which this object is available.
+        Access raises AttributeError if there is no object path, or more than
+        one object path.
+
+        Changed in 0.82.0: AttributeError can be raised.
+        """
+        if self._object_path is _MANY:
+            raise AttributeError('Object %r has more than one object path: '
+                                 'use Object.locations instead' % self)
+        elif self._object_path is None:
+            raise AttributeError('Object %r has no object path yet' % self)
+        else:
+            return self._object_path
+
+    @property
+    def connection(self):
+        """The Connection on which this object is available.
+        Access raises AttributeError if there is no Connection, or more than
+        one Connection.
+
+        Changed in 0.82.0: AttributeError can be raised.
+        """
+        if self._connection is _MANY:
+            raise AttributeError('Object %r is on more than one Connection: '
+                                 'use Object.locations instead' % self)
+        elif self._connection is None:
+            raise AttributeError('Object %r has no Connection yet' % self)
+        else:
+            return self._connection
+
+    @property
+    def locations(self):
+        """An iterable over tuples representing locations at which this
+        object is available.
+
+        Each tuple has at least two items, but may have more in future
+        versions of dbus-python, so do not rely on their exact length.
+        The first two items are the dbus.connection.Connection and the object
+        path.
+
+        :Since: 0.82.0
+        """
+        return iter(self._locations)
+
+    def add_to_connection(self, connection, path):
+        """Make this object accessible via the given D-Bus connection and
+        object path.
+
+        :Parameters:
+            `connection` : dbus.connection.Connection
+                Export the object on this connection. If the class attribute
+                SUPPORTS_MULTIPLE_CONNECTIONS is False (default), this object
+                can only be made available on one connection; if the class
+                attribute is set True by a subclass, the object can be made
+                available on more than one connection.
+
+            `path` : dbus.ObjectPath or other str
+                Place the object at this object path. If the class attribute
+                SUPPORTS_MULTIPLE_OBJECT_PATHS is False (default), this object
+                can only be made available at one object path; if the class
+                attribute is set True by a subclass, the object can be made
+                available with more than one object path.
+
+        :Raises ValueError: if the object's class attributes do not allow the
+            object to be exported in the desired way.
+        :Since: 0.82.0
+        """
+        if path == LOCAL_PATH:
+            raise ValueError('Objects may not be exported on the reserved '
+                             'path %s' % LOCAL_PATH)
+
+        self._locations_lock.acquire()
+        try:
+            if (self._connection is not None and
+                self._connection is not connection and
+                not self.SUPPORTS_MULTIPLE_CONNECTIONS):
+                raise ValueError('%r is already exported on '
+                                 'connection %r' % (self, self._connection))
+
+            if (self._object_path is not None and
+                not self.SUPPORTS_MULTIPLE_OBJECT_PATHS and
+                self._object_path != path):
+                raise ValueError('%r is already exported at object '
+                                 'path %s' % (self, self._object_path))
+
+            connection._register_object_path(path, self._message_cb,
+                                             self._unregister_cb,
+                                             self._fallback)
+
+            if self._connection is None:
+                self._connection = connection
+            elif self._connection is not connection:
+                self._connection = _MANY
+
+            if self._object_path is None:
+                self._object_path = path
+            elif self._object_path != path:
+                self._object_path = _MANY
+
+            self._locations.append((connection, path, self._fallback))
+        finally:
+            self._locations_lock.release()
+
+    def remove_from_connection(self, connection=None, path=None):
+        """Make this object inaccessible via the given D-Bus connection
+        and object path. If no connection or path is specified,
+        the object ceases to be accessible via any connection or path.
+
+        It's not currently possible to export an object on more than one
+        connection or with more than one object-path, but this will be
+        supported in future.
+
+        :Parameters:
+            `connection` : dbus.connection.Connection or None
+                Only remove the object from this Connection. If None,
+                remove from all Connections on which it's exported.
+            `path` : dbus.ObjectPath or other str, or None
+                Only remove the object from this object path. If None,
+                remove from all object paths.
+        :Raises LookupError:
+            if the object was not exported on the requested connection
+            or path, or (if both are None) was not exported at all.
+        :Since: 0.81.1
+        """
+        self._locations_lock.acquire()
+        try:
+            if self._object_path is None or self._connection is None:
+                raise LookupError('%r is not exported' % self)
+
+            if connection is not None or path is not None:
+                dropped = []
+                for location in self._locations:
+                    if ((connection is None or location[0] is connection) and
+                        (path is None or location[1] == path)):
+                        dropped.append(location)
+            else:
+                dropped = self._locations
+                self._locations = []
+
+            if not dropped:
+                raise LookupError('%r is not exported at a location matching '
+                                  '(%r,%r)' % (self, connection, path))
+
+            for location in dropped:
+                try:
+                    location[0]._unregister_object_path(location[1])
+                except LookupError:
+                    pass
+                if self._locations:
+                    try:
+                        self._locations.remove(location)
+                    except ValueError:
+                        pass
+        finally:
+            self._locations_lock.release()
 
     def _unregister_cb(self, connection):
-        _logger.info('Unregistering exported object %r', self)
+        # there's not really enough information to do anything useful here
+        _logger.info('Unregistering exported object %r from some path '
+                     'on %r', self, connection)
 
     def _message_cb(self, connection, message):
         try:
@@ -435,7 +638,7 @@ class Object(Interface):
             keywords = {}
 
             if parent_method._dbus_out_signature is not None:
-                signature = _dbus_bindings.Signature(parent_method._dbus_out_signature)
+                signature = Signature(parent_method._dbus_out_signature)
             else:
                 signature = None
 
@@ -445,9 +648,40 @@ class Object(Interface):
                 keywords[return_callback] = lambda *retval: _method_reply_return(connection, message, method_name, signature, *retval)
                 keywords[error_callback] = lambda exception: _method_reply_error(connection, message, exception)
 
-            # include the sender if desired
+            # include the sender etc. if desired
             if parent_method._dbus_sender_keyword:
                 keywords[parent_method._dbus_sender_keyword] = message.get_sender()
+            if parent_method._dbus_path_keyword:
+                keywords[parent_method._dbus_path_keyword] = message.get_path()
+            if parent_method._dbus_rel_path_keyword:
+                path = message.get_path()
+                rel_path = path
+                for exp in self._locations:
+                    # pathological case: if we're exported in two places,
+                    # one of which is a subtree of the other, then pick the
+                    # subtree by preference (i.e. minimize the length of
+                    # rel_path)
+                    if exp[0] is connection:
+                        if path == exp[1]:
+                            rel_path = '/'
+                            break
+                        if exp[1] == '/':
+                            # we already have rel_path == path at the beginning
+                            continue
+                        if path.startswith(exp[1] + '/'):
+                            # yes we're in this exported subtree
+                            suffix = path[len(exp[1]):]
+                            if len(suffix) < len(rel_path):
+                                rel_path = suffix
+                rel_path = ObjectPath(rel_path)
+                keywords[parent_method._dbus_rel_path_keyword] = rel_path
+
+            if parent_method._dbus_destination_keyword:
+                keywords[parent_method._dbus_destination_keyword] = message.get_destination()
+            if parent_method._dbus_message_keyword:
+                keywords[parent_method._dbus_message_keyword] = message
+            if parent_method._dbus_connection_keyword:
+                keywords[parent_method._dbus_connection_keyword] = connection
 
             # call method
             retval = candidate_method(self, *args, **keywords)
@@ -482,8 +716,14 @@ class Object(Interface):
 
             # no signature, so just turn the return into a tuple and send it as normal
             else:
-                if retval == None:
+                if retval is None:
                     retval = ()
+                elif (isinstance(retval, tuple)
+                      and not isinstance(retval, Struct)):
+                # If the return is a tuple that is not a Struct, we use it
+                # as-is on the assumption that there are multiple return
+                # values - this is the usual Python idiom. (fd.o #10174)
+                    pass
                 else:
                     retval = (retval,)
 
@@ -492,13 +732,14 @@ class Object(Interface):
             # send error reply
             _method_reply_error(connection, message, exception)
 
-    @method('org.freedesktop.DBus.Introspectable', in_signature='', out_signature='s')
-    def Introspect(self):
+    @method(INTROSPECTABLE_IFACE, in_signature='', out_signature='s',
+            path_keyword='object_path', connection_keyword='connection')
+    def Introspect(self, object_path, connection):
         """Return a string of XML encoding this object's supported interfaces,
         methods and signals.
         """
-        reflection_data = '<!DOCTYPE node PUBLIC "-//freedesktop//DTD D-BUS Object Introspection 1.0//EN" "http://www.freedesktop.org/standards/dbus/1.0/introspect.dtd">\n'
-        reflection_data += '<node name="%s">\n' % (self._object_path)
+        reflection_data = _dbus_bindings.DBUS_INTROSPECT_1_0_XML_DOCTYPE_DECL_NODE
+        reflection_data += '<node name="%s">\n' % object_path
 
         interfaces = self._dbus_class_table[self.__class__.__module__ + '.' + self.__class__.__name__]
         for (name, funcs) in interfaces.iteritems():
@@ -512,11 +753,62 @@ class Object(Interface):
 
             reflection_data += '  </interface>\n'
 
+        for name in connection.list_exported_child_objects(object_path):
+            reflection_data += '  <node name="%s"/>\n' % name
+
         reflection_data += '</node>\n'
 
         return reflection_data
 
     def __repr__(self):
-        return '<dbus.service.Object %s on %r at %#x>' % (self._object_path, self._name, id(self))
+        where = ''
+        if (self._object_path is not _MANY
+            and self._object_path is not None):
+            where = ' at %s' % self._object_path
+        return '<%s.%s%s at %#x>' % (self.__class__.__module__,
+                                   self.__class__.__name__, where,
+                                   id(self))
     __str__ = __repr__
 
+class FallbackObject(Object):
+    """An object that implements an entire subtree of the object-path
+    tree.
+
+    :Since: 0.82.0
+    """
+
+    SUPPORTS_MULTIPLE_OBJECT_PATHS = True
+
+    def __init__(self, conn=None, object_path=None):
+        """Constructor.
+
+        Note that the superclass' ``bus_name`` __init__ argument is not
+        supported here.
+
+        :Parameters:
+            `conn` : dbus.connection.Connection or None
+                The connection on which to export this object. If this is not
+                None, an `object_path` must also be provided.
+
+                If None, the object is not initially available on any
+                Connection.
+
+            `object_path` : str or None
+                A D-Bus object path at which to make this Object available
+                immediately. If this is not None, a `conn` must also be
+                provided.
+
+                This object will implements all object-paths in the subtree
+                starting at this object-path, except where a more specific
+                object has been added.
+        """
+        super(FallbackObject, self).__init__()
+        self._fallback = True
+
+        if conn is None:
+            if object_path is not None:
+                raise TypeError('If object_path is given, conn is required')
+        elif object_path is None:
+            raise TypeError('If conn is given, object_path is required')
+        else:
+            self.add_to_connection(conn, object_path)
